from __future__ import annotations


class WorkerGroup:
    APP_UPGRADE = "app_upgrade"
    BULK_PASTE = "bulk_paste"
    BULK_RENAME = "bulk_rename"
    BUNDLE_INSTALL = "bundle_install"
    CREATE_ARCHIVE = "create_archive"
    DEFAULT_SCRIPTS_NAMESPACE = "default_scripts_namespace"
    DEFAULT_SCRIPTS_UPDATE = "default_scripts_update"
    DELETE_PATH = "delete_path"
    DELETE_PATHS = "delete_paths"
    DIRECTORY_LISTING = "directory_listing"
    EXTRACT_ARCHIVE = "extract_archive"
    FILE_INFO = "file_info"
    MONDAY_SYNC = "monday_sync"
    SCRIPT_ABORT = "script_abort"
    SCRIPT_SNAPSHOT = "script_snapshot"
    SCRIPT_UPDATE_CHECK = "script_update_check"
    SCRIPTS = "scripts"
    SCRIPTS_REFRESH = "scripts_refresh"
    UPDATE_CHECK = "update_check"
    WATCHER_SNAPSHOT = "watcher_snapshot"
